/****************************************************************************
*  FILE:  main.cpp                                                          *
*  DESC:  Main entry point for tazman-listener.  Listens for TZSP stream    *
*         and outputs PCAP files.                                           *
*  TO-DO:                                                                   *
*     -Move into version control                                            *
*     -Add proper Tazman header parsing (when needed)                       *
*  CHANGELOG:                                                               *
*  09-26-17/BCG                                                             *
*     Initial effort using MyHoneyPot (MyHP) code as a starting point       *
*  10-04-17/BCG                                                             *
*     First alpha-level code deployed on Moloch server                      *
*  10-06-17/BCG  v1.0                                                       *
*     First release-level code pushed to Moloch server                      *
****************************************************************************/

/*****************
*  DEPENDENCIES  *
*****************/
#include <unistd.h>           // for getopt()
#include <string.h>           // for memset()
#include <netdb.h>            // for struct addrinfo
#include <sys/time.h>         // for gettimeofday()
#include <dirent.h>           // for openddir()
#include <pwd.h>              // for passwd struct, etc
#include <net/if.h>           // for ifreq struct
#include <sys/ioctl.h>        // for ioctl()
#include "logging.h"          // our own handy logging code


/**************
*  CONSTANTS  *
**************/
const int PACKET_BUF_SIZE             = 65535;  // default packet read buffer size
const int DEFAULT_PCAP_NUM_PACKETS    = 10000;  // default #packets per output pcap file
const int DEFAULT_NETWORK_PORT        = 37008;  // default network interface to receive tzsp


/************
*  GLOBALS  *
************/
int  PCAP_NUM_PACKETS;     // # packets per PCAP file (override with -c)
int  NETWORK_PORT;         // default port number to listen on
char PROGRAM_NAME[255];    // our name-improves readability when we abend
char OUTFILE_PREFIX[255];  // prefix to prepend to pcap outfiles
char RUNAS_USER[255];      // to change user profile/priviledges at runtime
int  ECHO_SOCKET_FD = 0;   // sockfd to echo/export recieved packets


/************
*  STRUCTS  *
************/
// wiki.wireshark.org/Development/LibpcapFileFormat
struct pcap_packet_hdr_s
{
   uint32_t ts_sec;
   uint32_t ts_usec;
   uint32_t incl_len;
   uint32_t orig_len;
};

// wiki.wireshark.org/Development/LibpcapFileFormat
struct pcap_file_hdr_s
{
   uint32_t magic_number  = 0xa1b2c3d4;
   uint16_t version_major = 2;
   uint16_t version_minor = 4;
   int32_t  thiszone      = 0;
   uint32_t sigfigs       = 0;
   uint32_t snaplen       = 65535;
   uint32_t network       = 1;
};


/*******************************************************************
*  FUNCTION DECLARATIONS                                           *
*  Allows us to put helper functions further down for readability  *
*******************************************************************/
void showHelpAbend();                             // responsible for displaying options and abending
void setSafeDefaults();                           // responsible for setting globals to safe defaults
void parseArgs(int, char**);                      // responsible for reading argv and setting globals
void startEcho();                                 // responsible for opening the echo/export socket
void startListener();                             // responsible for starting the libpcap socket listener
inline int  open_pcap();                          // open a new pcap output file and write pcap file header
inline int  write_pcap_packet(int, char*, int);   // writes pcap packet header and packet data to pcap output file


/****************************************************************************
*  main()                                                                   *
*  Entry point for application.  Responsible for parsing args, handling any *
*  dependencies, and starting the socket listener.                          *
****************************************************************************/
int main(int argc, char** argv)
{
   setSafeDefaults();
   parseArgs(argc, argv);
   startEcho();
   startListener();
   return 0;

} // end main()


/*************************************************
*  showHelpAbend()                               *
*  Terminal function used to display usage help  *
*  and exit with abnormal exit value of 1        *
*************************************************/
void showHelpAbend()
{
   // invoked name tot readily available :-(
   const char* str =
"***USAGE: %s [options]\n \
    Available Options:\n \
       -c <#packets>  - num packets per pcap file \n \
       -h             - show this help and exit \n \
       -o <path>      - prepend to pcap output file path \n \
       -p <portnum>   - set UDP port to listen on (default=37008)\n \
       -u <uname>     - run as <uname> user (must start as root)\n\n \
";

   printf(str, PROGRAM_NAME);
   exit(1);

} // end showHelpAbend()


/**********************************************************
*  setSafeDefaults()                                      *
*  Responsible for setting our globalss to safe defaults  *
**********************************************************/
void setSafeDefaults()
{
   PCAP_NUM_PACKETS = DEFAULT_PCAP_NUM_PACKETS;
   NETWORK_PORT = DEFAULT_NETWORK_PORT;
   strcpy(OUTFILE_PREFIX, "./");
   strcpy(RUNAS_USER, "");

} // end setSafeDefaults()


/*******************************************************
*  parseArgs()                                         *
*  Function used to parse/process argv and initialize  *
*  global settings accordingly.                        *
*******************************************************/
void parseArgs(int argc, char** argv)
{
   strcpy(PROGRAM_NAME, argv[0]);
   int opt;

   while((opt = getopt(argc, argv, "c:o:p:u:h")) != -1)
   {
      switch(opt)
      {
         case 'c':
            PCAP_NUM_PACKETS = atoi(optarg);
            LOG("Packets per PCAP file set by caller to '%s' (-c #packets)", optarg);
            break;
         case 'o':          // caller is setting a pcap output prefix
            strcpy(OUTFILE_PREFIX, optarg);
            LOG("PCAP output prefix set by caller to '%s' (-o <iface>)", optarg);
            break;
         case 'p':          // caller is specifying a listening port
            NETWORK_PORT = atoi(optarg);
            LOG("Port set by caller to '%s' (-p <portnum>)", optarg);
            break;
         case 'u':
            strcpy(RUNAS_USER, optarg);
            LOG("Runas user set to '%s' (-u <uname>)", optarg);
            break;
         case 'h':          // caller is asking for help
         default:           // any unknown args
            showHelpAbend();
      } // end switch()
   } // end while()

   // if we made it to here, everything is syntactically correct-time to sanitize
   if(PCAP_NUM_PACKETS <= 0)
      LOG("Invalid Number of Packets per PCAP File (option -c)!!", LOGLEVEL_ABEND);

   if(NETWORK_PORT <= 0 || NETWORK_PORT > 65535)
      LOG("Invalid Network Port (option -p)!!", LOGLEVEL_ABEND);

   DIR* dir = opendir(OUTFILE_PREFIX);
   if(dir)
      closedir(dir);
   else
      LOG("Invalid PCAP Output Directory (option -o)!!", LOGLEVEL_ABEND);
   if('/' != OUTFILE_PREFIX[strlen(OUTFILE_PREFIX)-1])
      LOG("PCAP Output Directory (option -o) Must End With '/' or Bad Things Can Happen!!", LOGLEVEL_ABEND);

   if(strlen(RUNAS_USER) > 0)
   {
      struct passwd* pwd = getpwnam(RUNAS_USER);
      if(pwd == NULL)
         LOG("RUNAS_USER '%s' not found!", RUNAS_USER, LOGLEVEL_ABEND);
      if(-1 ==setuid(pwd->pw_uid))
         LOG("Unable to run as user '%s' (requires root)", RUNAS_USER, LOGLEVEL_ABEND);

   } // end if()

} // end parseArgs()


//XXXTODO
void startEcho()
{
   int sd = socket(AF_PACKET, SOCK_RAW, IPPROTO_RAW);
   if(sd < 0)
      LOG("Unable to open raw socket for packet echo/export!", LOGLEVEL_ABEND);
   ECHO_SOCKET_FD = sd;

   /*
   struct ifreq ifr;
   memset(&ifr, 0, sizeof(ifr));
   strncpy(ifr.ifr_name, "eth1", IFNAMSIZ-1);

   if((ioctl(sd, SIOCGIFINDEX, &ifr)) < 0)
      LOG("Unable to find echo/export interface (ioctl error)!", LOGLEVEL_ABEND);
   */


} // end startEcho()


/********************************************************
*  startListener()                                      *
*  Responsible for opening network socket, listening    *
*  for tazman packets, and writing packets out to PCAP  *
********************************************************/
void startListener()
{
   int sfd = socket(AF_INET, SOCK_DGRAM, 0);
   if(sfd < 0)
      LOG("Unable to open socket on network interface!", LOGLEVEL_ABEND);

   struct sockaddr_in saddr;
   bzero((char*)&saddr, sizeof(saddr));
   saddr.sin_family = AF_INET;
   saddr.sin_port = htons(NETWORK_PORT);
   saddr.sin_addr.s_addr = htonl(INADDR_ANY);

   if(bind(sfd, (struct sockaddr*)&saddr, sizeof(saddr)) < -1)
      LOG("Unable to bind to listening socket!", LOGLEVEL_ABEND);

   struct sockaddr_in paddr;
   socklen_t paddr_len = sizeof(sockaddr_in);
   int nread;
   char buf[PACKET_BUF_SIZE];
   int pct = 0;
   int pfd = open_pcap();
   while(true)
   {
      nread = recvfrom(sfd, buf, PACKET_BUF_SIZE, 0, (struct sockaddr*) &paddr, &paddr_len);
      if(nread < 0)
         LOG("Unable to read from listening socket!", LOGLEVEL_ABEND);
      else if(nread < 5)
         continue;   // protect against truncated packets (necessary?)

      // note: Mikrotik seems to just use fixed 5-byte tzsp headers-might change in the future
      // makes truncating easy/fast for now though
      if(write_pcap_packet(pfd, buf+5, nread-5) <= 0)
        LOG("Unable to write packet to pcap file!", LOGLEVEL_ERROR); // treat as non-fatal (for now)

      if(ECHO_SOCKET_FD > 0)
      {
         write(ECHO_SOCKET_FD, buf+5, nread-5);
      }

      // time to start a new pcap?
      if(++pct > PCAP_NUM_PACKETS)
      {
         close(pfd);
         pfd = open_pcap();
         pct = 0;
      } // end if()

   } // end while(true)

} // end startListener()


/********************************************************
*  open_pcap()                                          *
*  Opend a new pcap file and writes a pcap file header  *
********************************************************/
inline int open_pcap()
{
   // get a timestamp
   time_t t = time(NULL);         // declare a time_t instance with system time
   struct tm* ts = localtime(&t); // tm struct containing localized time

   // open a timestamped pcap file for writing
   char fname[255];
   sprintf(fname, "%s\%04d%02d%02d_%02d%02d%02d.%s",
                  OUTFILE_PREFIX,
                  1900+ts->tm_year,
                  ts->tm_mon+1,
                  ts->tm_mday,
                  ts->tm_hour,
                  ts->tm_min,
                  ts->tm_sec,
                  "XXXXXX.pcap");  // throw in 6 random chars (~48 bits) to ensure uniqueness
   int fd = mkstemps(fname,5);     // 5 is the number of chars AFTER the hex wildcard
   if(fd < 0)
      LOG("Unable to open PCAP file for writing!!", LOGLEVEL_ABEND);

   LOG("writing packets out to %s\n", fname);

   // write out a valid pcap file header
   const struct pcap_file_hdr_s hdr;
   if(-1 == write(fd, &hdr, sizeof(pcap_file_hdr_s)))
      LOG("Unable to write PCAP file header!!", LOGLEVEL_ABEND);

   // return fd to caller for further use
   return fd;

} // end open_pcap()


/********************************************************
*  write_pcap_packet(int fd, char* pkt, int pktlen)     *
*  Write given packet out to the given file descriptor  *
*  with an appropriate pcap packet header.              *
********************************************************/
inline int write_pcap_packet(int fd, char* pkt, int pktlen)
{
   struct pcap_packet_hdr_s pkthdr;
   struct timeval ts;

   // populate and write-out the pcap packet header...
   gettimeofday(&ts, NULL);
   pkthdr.incl_len = pktlen;
   pkthdr.orig_len = pktlen;
   pkthdr.ts_sec   = ts.tv_sec;
   pkthdr.ts_usec  = ts.tv_usec;
   write(fd, &pkthdr, sizeof(pcap_packet_hdr_s));

   // write out the packet data and return result
   return write(fd, pkt, pktlen);

} // end write_pcap_packet()


